import asyncio

import asynctest
import pytest
import aioredis

import fakeredis.aioredis


class TestFakeCommands(asynctest.TestCase):
    async def setUp(self):
        self.redis = await fakeredis.aioredis.create_redis_pool()

    async def tearDown(self):
        self.redis.close()
        await self.redis.wait_closed()

    async def test_ping(self):
        pong = await self.redis.ping()
        assert pong == b'PONG'

    async def test_types(self):
        await self.redis.hmset_dict('hash', key1='value1', key2='value2', key3=123)
        result = await self.redis.hgetall('hash', encoding='utf-8')
        assert result == {
            'key1': 'value1',
            'key2': 'value2',
            'key3': '123'
        }

    async def test_transaction(self):
        tr = self.redis.multi_exec()
        tr.set('key1', 'value1')
        tr.set('key2', 'value2')
        ok1, ok2 = await tr.execute()
        assert ok1
        assert ok2
        result = await self.redis.get('key1')
        assert result == b'value1'

    async def test_transaction_fail(self):
        # ensure that the WATCH applies to the same connection as the MULTI/EXEC.
        await self.redis.set('foo', '1')
        with await self.redis as r:
            await r.watch('foo')
            await self.redis.set('foo', '2')    # Different connection
            tr = r.multi_exec()
            tr.get('foo')
            with pytest.raises(aioredis.MultiExecError):
                await tr.execute()

    async def test_pubsub(self):
        ch, = await self.redis.subscribe('channel')
        queue = asyncio.Queue()

        async def reader(channel):
            async for message in ch.iter():
                queue.put_nowait(message)

        task = self.loop.create_task(reader(ch))
        await self.redis.publish('channel', 'message1')
        await self.redis.publish('channel', 'message2')
        result1 = await queue.get()
        result2 = await queue.get()
        assert result1 == b'message1'
        assert result2 == b'message2'
        ch.close()
        await task

    async def test_blocking_ready(self):
        """Blocking command which does not need to block."""
        self.redis.rpush('list', 'x')
        with await self.redis as r:
            result = await r.blpop('list', timeout=1)
        assert result == [b'list', b'x']

    @pytest.mark.slow
    async def test_blocking_timeout(self):
        """Blocking command that times out without completing."""
        with await self.redis as r:
            result = await r.blpop('missing', timeout=1)
        assert result is None

    @pytest.mark.slow
    async def test_blocking_unblock(self):
        """Blocking command that gets unblocked after some time."""
        async def unblock():
            await asyncio.sleep(0.1)
            await self.redis.rpush('list', 'y')

        with await self.redis as r:
            task = self.loop.create_task(unblock())
            result = await r.blpop('list', timeout=1)
        assert result == [b'list', b'y']
        await task

    @pytest.mark.slow
    async def test_blocking_pipeline(self):
        """Blocking command with another command issued behind it."""
        with await self.redis as r:   # Ensure commands use same connection
            await r.set('foo', 'bar')
            fut = asyncio.ensure_future(r.blpop('list', timeout=1))
            assert (await r.get('foo')) == b'bar'
            assert (await fut) is None

    async def test_wrongtype_error(self):
        await self.redis.set('foo', 'bar')
        with pytest.raises(aioredis.ReplyError) as excinfo:
            await self.redis.rpush('foo', 'baz')
        assert str(excinfo.value).startswith('WRONGTYPE ')

    async def test_syntax_error(self):
        with pytest.raises(aioredis.ReplyError) as excinfo:
            await self.redis.execute('get')
        assert str(excinfo.value) == "ERR wrong number of arguments for 'get' command"

    async def test_no_script_error(self):
        with pytest.raises(aioredis.ReplyError) as excinfo:
            await self.redis.evalsha('0123456789abcdef0123456789abcdef')
        assert str(excinfo.value).startswith('NOSCRIPT ')

    async def test_failed_script_error(self):
        await self.redis.set('foo', 'bar')
        with pytest.raises(aioredis.ReplyError) as excinfo:
            await self.redis.eval(
                'return redis.call("ZCOUNT", KEYS[1])', ['foo'])
        assert str(excinfo.value).startswith('ERR Error running script')


class TestRealCommands(TestFakeCommands):
    async def setUp(self):
        try:
            self.redis = await aioredis.create_redis_pool('redis://localhost')
            await self.redis.flushall()
        except ConnectionRefusedError:
            pytest.skip('redis is not running')

    async def tearDown(self):
        await self.redis.flushall()
        await super().tearDown()



import asyncio
import sys
import warnings

import aioredis
import async_timeout

from . import _server


class FakeSocket(_server.FakeSocket):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.responses = asyncio.Queue()

    def put_response(self, msg):
        self.responses.put_nowait(msg)

    def _decode_error(self, error):
        return aioredis.ReplyError(error.value)

    async def _async_blocking(self, timeout, func, event, callback):
        try:
            result = None
            with async_timeout.timeout(timeout if timeout else None):
                while True:
                    await event.wait()
                    event.clear()
                    # This is a coroutine outside the normal control flow that
                    # locks the server, so we have to take our own lock.
                    with self._server.lock:
                        ret = func(False)
                        if ret is not None:
                            result = self._decode_result(ret)
                            self.put_response(result)
                            break
        except asyncio.TimeoutError:
            result = None
        finally:
            with self._server.lock:
                self._db.remove_change_callback(callback)
            self.put_response(result)
            self.resume()

    def _blocking(self, timeout, func):
        loop = asyncio.get_event_loop()
        ret = func(True)
        if ret is not None or self._in_transaction:
            return ret
        event = asyncio.Event()

        def callback():
            loop.call_soon_threadsafe(event.set)
        self._db.add_change_callback(callback)
        self.pause()
        loop.create_task(self._async_blocking(timeout, func, event, callback))
        return _server.NoResponse()


class FakeReader:
    """Re-implementation of aioredis.stream.StreamReader.

    It does not use a socket, but instead provides a queue that feeds
    `readobj`.
    """

    def __init__(self, socket):
        self._socket = socket

    def set_parser(self, parser):
        pass       # No parser needed, we get already-parsed data

    async def readobj(self):
        if self._socket.responses is None:
            raise asyncio.CancelledError
        result = await self._socket.responses.get()
        return result

    def at_eof(self):
        return self._socket.responses is None

    def feed_obj(self, obj):
        self._queue.put_nowait(obj)


class FakeWriter:
    """Replaces a StreamWriter for an aioredis connection."""

    def __init__(self, socket):
        self.transport = socket       # So that aioredis can call writer.transport.close()

    def write(self, data):
        self.transport.sendall(data)


class FakeConnectionsPool(aioredis.ConnectionsPool):
    def __init__(self, server=None, db=None, password=None, encoding=None,
                 *, minsize, maxsize, ssl=None, parser=None,
                 create_connection_timeout=None,
                 connection_cls=None,
                 loop=None):
        super().__init__('fakeredis',
                         db=db,
                         password=password,
                         encoding=encoding,
                         minsize=minsize,
                         maxsize=maxsize,
                         ssl=ssl,
                         parser=parser,
                         create_connection_timeout=create_connection_timeout,
                         connection_cls=connection_cls,
                         loop=loop)
        if server is None:
            server = _server.FakeServer()
        self._server = server

    def _create_new_connection(self, address):
        # TODO: what does address do here? Might just be for sentinel?
        return create_connection(self._server,
                                 db=self._db,
                                 password=self._password,
                                 ssl=self._ssl,
                                 encoding=self._encoding,
                                 parser=self._parser_class,
                                 timeout=self._create_connection_timeout,
                                 connection_cls=self._connection_cls,
                                 )


async def create_connection(server=None, *, db=None, password=None, ssl=None,
                            encoding=None, parser=None, loop=None,
                            timeout=None, connection_cls=None):
    # This is mostly copied from aioredis.connection.create_connection
    if timeout is not None and timeout <= 0:
        raise ValueError("Timeout has to be None or a number greater than 0")

    if connection_cls:
        assert issubclass(connection_cls, aioredis.abc.AbcConnection),\
                "connection_class does not meet the AbcConnection contract"
        cls = connection_cls
    else:
        cls = aioredis.connection.RedisConnection

    if loop is not None and sys.version_info >= (3, 8, 0):
        warnings.warn("The loop argument is deprecated",
                      DeprecationWarning)

    if server is None:
        server = _server.FakeServer()
    socket = FakeSocket(server)
    reader = FakeReader(socket)
    writer = FakeWriter(socket)
    conn = cls(reader, writer, encoding=encoding,
               address='fakeredis', parser=parser)

    try:
        if password is not None:
            await conn.auth(password)
        if db is not None:
            await conn.select(db)
    except Exception:
        conn.close()
        await conn.wait_closed()
        raise
    return conn


async def create_redis(server=None, *, db=None, password=None, ssl=None,
                       encoding=None, commands_factory=aioredis.Redis,
                       parser=None, timeout=None,
                       connection_cls=None, loop=None):
    conn = await create_connection(server, db=db,
                                   password=password,
                                   ssl=ssl,
                                   encoding=encoding,
                                   parser=parser,
                                   timeout=timeout,
                                   connection_cls=connection_cls,
                                   loop=loop)
    return commands_factory(conn)


async def create_pool(server=None, *, db=None, password=None, ssl=None,
                      encoding=None, minsize=1, maxsize=10,
                      parser=None, loop=None, create_connection_timeout=None,
                      pool_cls=None, connection_cls=None):
    # Mostly copied from aioredis.pool.create_pool.
    if pool_cls:
        assert issubclass(pool_cls, aioredis.AbcPool),\
                "pool_class does not meet the AbcPool contract"
        cls = pool_cls
    else:
        cls = FakeConnectionsPool

    pool = cls(server, db, password, encoding,
               minsize=minsize, maxsize=maxsize,
               ssl=ssl, parser=parser,
               create_connection_timeout=create_connection_timeout,
               connection_cls=connection_cls,
               loop=loop)
    try:
        await pool._fill_free(override_min=False)
    except Exception:
        pool.close()
        await pool.wait_closed()
        raise
    return pool


async def create_redis_pool(server=None, *, db=None, password=None, ssl=None,
                            encoding=None, commands_factory=aioredis.Redis,
                            minsize=1, maxsize=10, parser=None,
                            timeout=None, pool_cls=None,
                            connection_cls=None, loop=None):
    pool = await create_pool(server, db=db,
                             password=password,
                             ssl=ssl,
                             encoding=encoding,
                             minsize=minsize,
                             maxsize=maxsize,
                             parser=parser,
                             create_connection_timeout=timeout,
                             pool_cls=pool_cls,
                             connection_cls=connection_cls,
                             loop=loop)
    return commands_factory(pool)
